# -*- coding: utf-8 -*-
"""

LETKF plot suite

Plots hourly output data from MONARCH LETKF, compute misfits,
innovation statistics, model fields, analysis increments
and ensemble standard deviation

TODO:
    - [ ] use experiment namelist to set script parameters
    - [ ] generalize instrument type (only VIIRS DOD currently)
"""

import os
import numpy as np
from netCDF4 import Dataset
import matplotlib.pyplot as plt
from plotters import plot2Dmap
from datetime import datetime as dt
from datetime import timedelta as td
from scipy.interpolate import griddata
import copy
import subprocess
import pandas as pd
import xarray as xr
from plotters import *
from argparse import ArgumentParser, RawDescriptionHelpFormatter

def readghostvis(date):
    
    archivedir = '/home/eemili/data'
    
    filename = os.path.join(archivedir,'vdist_%s.nc' % (date.strftime('%Y%m')))
    
    obsdata = xr.open_dataset(filename)
    
    return obsdata

def loopovertime(opt):
    
    # T03 is for dustclim starting at 00 UTC of the previous day
    # firstdate = dt.fromisoformat('2018-11-24T03')
    # T12 is for oper forecast starting at 12 UTC of the previous day
    firstdate = dt.fromisoformat('%sT12' % (opt.date))
    # firstdate = dt.fromisoformat('2024-07-12T00')
    ndays = int(opt.ndays)
    dates = [firstdate + i*td(days=1) for i in range(ndays)]

    fcstep = td(hours=3)
    # fcperiod = td(hours=60)
    fcperiod = td(hours=72)
    # First step to start making the plots
    # first_step = 4
    first_step = 2
    mvar = 'od550du'
    # mvar = 'sconcdu'
    # isregional = False
    isregional = True
    # MONARCH grid information (used to rotate back observations lats,lons)
    cenlon, cenlat = 20., 35.
    # cenlon is tph0d, cenlat is tlm0d
    # exppath = '/esarchive/exp/monarch/a60c'
    exppath = '/esarchive/exp/monarch/a2in'
    mpath = os.path.join(exppath,'regional/3hourly')
    
    pltbounds = {'LatMin':-11,'LatMax':70,'LonMin':-62,'LonMax':102}
    latbnd = (pltbounds['LatMin'],pltbounds['LatMax'])
    lonbnd = (pltbounds['LonMin'],pltbounds['LonMax'])
    # pltbounds = {'LatMin':15,'LatMax':55,'LonMin':-10,'LonMax':50}
    # pltbounds = {'LatMin':-90,'LatMax':90,'LonMin':-180,'LonMax':180}
    maprange = {'od550du':{'max':1.0,'min':0.},'sconcdu':{'max':0.6e-6,'min':0.}}
    mapdiffrange = {'od550du':{'max':0.5,'min':-0.5},'sconcdu':{'max':0.2e-6,'min':-0.2e-6}}
    mapstdrange = {'od550du':{'max':0.5,'min':0.},'sconcdu':{'max':0.4e-6,'min':0.}}
    maplowdiffrange =  {'od550du':{'max':0.3,'min':-0.3},'sconcdu':{'max':0.1e-6,'min':-0.1e-6}}
    histrange = {'od550du':{'max':1.5,'min':0.}}
    histdiffrange = {'od550du':{'max':1.5,'min':-1.5}}
    histlowdiffrange = {'od550du':{'max':1.,'min':-1.}}
    markersize = 1.0
    
    # if not os.path.exists(outdir): os.makedirs(outdir)
    # if not os.path.exists(framedir): os.makedirs(framedir)
    # if not os.path.exists(gifdir): os.makedirs(gifdir)
    for idate, date in enumerate(dates):

        mfile = os.path.join(mpath,'%s' % (mvar),'%s_%03i_%s.nc' % (mvar,0,date.strftime('%Y%m%d')))
        outdir = os.path.join(exppath,'plots',date.strftime('%Y%m%d'))
        framedir = os.path.join(outdir,'frames')
        gifdir = os.path.join(outdir,'gifs')
        
        inidate = date
        fcsteps = [inidate+i*fcstep for i in range(int(fcperiod/fcstep))]

        # yall = obsdb.DefInstrument('VIIRS','VIIRS_AOD',wavelength=[550])
        nobs, y_ts, hofx_ts, fcdep_ts = [[] for i in range(4)]

        # Now loop on output frequency and do some plots
        for step in fcsteps[first_step:]:
            istep = int((step-inidate)/fcstep)
            print(step)
            print(istep)
            
            # Read and plot observations, departures and hofxbg. 
            # Object are reinitialized to avoid cumulating the obs for the moment
            
            visdata = readghostvis(step)
            
            sitenames, vislats, vislons = visdata['station_name'][:], visdata['latitude'][:], visdata['longitude'][:]
            
            mask = np.zeros(len(sitenames),dtype=bool)
            
            if latbnd is not None: 
                print("Selecting obs latitudes between ",latbnd[0]," and ",latbnd[1])
                latmsk = ~((vislats > latbnd[0]) & (vislats < latbnd[1]))
                if latmsk.any(): mask = mask | latmsk
            if lonbnd is not None:
                print("Selecting obs longitudes between ",lonbnd[0]," and ",lonbnd[1])
                lonmsk = ~((vislons > lonbnd[0]) & (vislons < lonbnd[1]))
                if lonmsk.any(): mask = mask | lonmsk
                
            sitenames = np.ma.array(sitenames[~mask])
            vislats = np.ma.array(vislats[~mask])
            vislons = np.ma.array(vislons[~mask])
            
            #vdist = visdata['vdist'].sel(time=slice(step-td(hours=1.5),step+td(hours=1.5)))[~mask].values
            vdist = np.ma.array(visdata['vdist'][:,:].sel(time=slice(step-td(hours=0.5),
                                                                                    step+td(hours=0.5)))[~mask].values.squeeze())
            oflag = np.ma.array(visdata['flag'].sel(time=slice(step-td(hours=0.5),step+td(hours=0.5)))[~mask].values.squeeze(),dtype=np.int16)
            qaflag = np.ma.array(visdata['qa'].sel(time=slice(step-td(hours=0.5),step+td(hours=0.5)))[~mask].values.squeeze(),dtype=np.int16)
            
            # Mask nan values
            vdist[np.isnan(vdist)]= np.ma.masked
            qaflag[np.isnan(qaflag)] = np.ma.masked # Not really needing this for now
            oflag[np.isnan(oflag)] = np.ma.masked
            # Mask sites with nan values
            vislats.mask = vdist.mask
            vislons.mask = vdist.mask
            
            myflag = np.ma.zeros(len(vdist),dtype=np.int8)
            # Flag all valid observations and mask the others
            myflag[oflag[:,0] == 0] = 1
            vdist[oflag[:,0] != 0] = np.ma.masked
            
            # Flag using weather flags (duststorm/sandstorm) local sand/dust contamination (182 to 187 plus haze 194)
            myflag[np.ma.any((oflag == 176) | (oflag == 177)
                      | (oflag == 182) | (oflag == 183) | (oflag == 184) | (oflag == 185) 
                      | (oflag == 186) | (oflag == 187), axis=1)] = 2
                       # | (oflag == 186) | (oflag == 187) | (oflag == 194), axis=1)] = 2
            
            # Now flag dust using visibility on top of previous flags
            myflag[(vdist > 2000) & (vdist <= 5000) & (myflag > 1)] = 3
            myflag[(vdist > 1000) & (vdist <= 2000) & (myflag > 1)] = 4
            myflag[(vdist <= 1000) & (myflag > 1)] = 5
            
            plotobshisto(myflag)
            
            # myflag[(vdist > 2000) & (vdist < 5000)] = 3
            # myflag[(vdist > 1000) & (vdist < 2000)] = 4
            # myflag[(vdist < 1000)] = 5
            
            # plotobsloc(vislats,vislons)
            
            # plotobsval(myflag,vislats,vislons,range={'min':0,'max':5},colormap="Set1_r")
            
            # plotobsval(vdist,vislats,vislons,range={'min':1e3,'max':10e3})
            # plotobshisto(vdist)
            
            vdist[myflag < 2] = np.ma.masked
            # plotobsval(vdist,vislats,vislons,range={'min':1e3,'max':10e3})
            # plotobshisto(vdist)
            
            # Forecast field
            mfile = os.path.join(mpath,'%s' % (mvar),'%s-%03i_%s.nc' % (mvar,0,inidate.strftime('%Y%m%d%H')))
            f = Dataset(mfile, 'r', format="NETCDF4")
            xfc = f[mvar][istep,0,:,:]
            
            if isregional:
                lat_grid = f['lat'][:]
                lon_grid = f['lon'][:]
            else:
                grid = np.meshgrid(f['lon'][:],f['lat'][:])
                lat_grid = grid[1]
                lon_grid = grid[0]
                
            f.close()
            
            plot2Dmap(xfc,lat_grid,lon_grid,range=maprange[mvar],
                    title=step.strftime('%Y-%m-%d-%H')+' FG\n Avg:%.2e Min:%.2e Max:%.2e' % (xfc.mean(),xfc.min(),xfc.max()),
                    plotregion=pltbounds)
                    # filename=os.path.join(framedir,'monarch-%s-%s-fc' % (mvar,step.strftime('%Y%m%d%H'))))
            
            # -------------------------------------------------------------------------
            # Interpolate forecast data at the observation locations
            # -------------------------------------------------------------------------
            
            interpolated_data = griddata(
                (lon_grid.ravel(), lat_grid.ravel()),  # Known points
                xfc.ravel(),  # Known data
                np.array((vislons, vislats)).T,  # Points to interpolate to
                method='linear'  # You can use 'linear', 'nearest', or other methods
            )
        
            xfc_int = np.ma.masked_invalid(interpolated_data)
            
            plotobsval(xfc_int,vislats,vislons,range={'min':0,'max':1.0},label='AOD')
            
            # Try a first scatter plot
            # np.random.seed(19680801)
            # data = [sorted(np.random.normal(0, std, 100)) for std in range(1, 5)]
            
            # fig, (ax1, ax2) = plt.subplots(nrows=1, ncols=2, figsize=(9, 4), sharey=True)
            
            # ax1.set_title('Default violin plot')
            # ax1.set_ylabel('Observed values')
            # ax1.violinplot(data)

            # plt.subplots_adjust(bottom=0.15, wspace=0.05)
            # plt.show()
            
            fig = plt.figure(figsize=(10,10),dpi=250)
            ax = fig.add_subplot(111)
            violins = [xfc_int[myflag == f].compressed() if len(xfc_int[myflag == f]) > 3 else np.zeros(5) for f in range(1,6)]
            ax.violinplot(violins,showmeans=False,
                  showmedians=True,quantiles=[[0.1,0.9] for i in range(5)])
            # plt.savefig('test')
            plt.show()
            plt.close(fig)
            
            # Compute stats and save them for a time series plot (not perfect solution since the obs
            # plots should be done by the class methods, but I haven't yet developed strategy to handle
            # multiple obs objects within the obs class)
            # nobs.append(y.nobs)
            # y_ts.append(y.Stats())
            # hofx_ts.append(hofx.Stats())
            # fcdep_ts.append(fcdep.Stats())
            # This can undo selections if needed
            # y.ResetSel()
            
            # return None

        # if useobs and mvar == 'od550du':
        #     # Plot the time series of observations here
        #     fig = plt.figure(figsize=(12,20),dpi=250)
            
        #     fig.autofmt_xdate()
            
        #     ax = fig.add_subplot(311)
        #     ax.set_title("Number of %s observations" % ('VIIRS'))
            
        #     ax.plot(fcsteps[first_step:],nobs,color='royalblue',\
        #             ls='-',lw=2,marker=None,label='Assimilated',markersize=4.0)
        #     # ax.xaxis.set_major_formatter(matplotlib.dates.DateFormatter('%d %b, %H'))
            
        #     ax.grid(True)   
        #     ax.axis(xmin=fcsteps[first_step],xmax=fcsteps[-1])
        #     ax.set_xlabel('Time (UTC)')
        #     ax.legend(loc=1)
            
        #     ax = fig.add_subplot(312)
        #     ax.set_title("%s statistics (%s)" % ('VIIRS DOD','unitless'))
        #     ax.set_ylabel("%s" % ('unitless'))
            
        #     ax.errorbar(fcsteps[first_step:],np.ma.array([v['550']['avg'] for v in hofx_ts]),
        #                 yerr=np.ma.array([v['550']['std'] for v in hofx_ts]),color='royalblue',
        #             ls=None,marker='o',label='hofx_fc',markersize=4.0,
        #             capsize=6, elinewidth=1.7, markeredgewidth=1.7) 
        #     ax.errorbar(fcsteps[first_step:],np.ma.array([v['550']['avg'] for v in y_ts]),
        #                 yerr=np.ma.array([v['550']['std'] for v in y_ts]),color='black',
        #             ls=None,marker='o',label='y',markersize=4.0,
        #             capsize=2, elinewidth=1., markeredgewidth=1.,alpha=0.7)
                        
        #     ax.grid(True)   
        #     ax.axis(xmin=fcsteps[first_step],xmax=fcsteps[-1])
        #     ax.set_xlabel('Time (UTC)')
        #     ax.legend(loc=1)
            
        #     ax = fig.add_subplot(313)
        #     ax.set_title("%s statistics (%s)" % ('VIIRS DOD','unitless'))
        #     ax.set_ylabel("%s" % ('unitless'))
            
        #     ax.errorbar(fcsteps[first_step:],np.ma.array([v['550']['avg'] for v in fcdep_ts]),
        #                 yerr=np.ma.array([v['550']['std'] for v in fcdep_ts]),color='royalblue',
        #             ls=None,marker='o',label='fcdep',markersize=4.0,
        #             capsize=6, elinewidth=1.7, markeredgewidth=1.7)
                
        #     ax.grid(True)
        #     ax.axis(xmin=fcsteps[first_step],xmax=fcsteps[-1])
        #     ax.set_xlabel('Time (UTC)')
        #     ax.legend(loc=1)
            
        #     plt.savefig(os.path.join(outdir,'VIIRS_%s_time_series.png' % (date[:-2])))
        #     plt.close(fig)

        #     np.savez(os.path.join(outdir,'VIIRS_%s_time_series.npz' % (date[:-2])),
        #             time=fcsteps[first_step:],
        #             nobs=np.array(nobs),
        #             y_avg=np.ma.array([v['550']['avg'] for v in y_ts]).filled(float(-1.0e38)),
        #             y_std=np.ma.array([v['550']['std'] for v in y_ts]).filled(float(-1.0e38)),
        #             hofxfc_avg=np.ma.array([v['550']['avg'] for v in hofx_ts]).filled(float(-1.0e38)),
        #             hofxfc_std=np.ma.array([v['550']['std'] for v in hofx_ts]).filled(float(-1.0e38)),
        #             fcdep_avg=np.ma.array([v['550']['avg'] for v in fcdep_ts]).filled(float(-1.0e38)),
        #             fcdep_std=np.ma.array([v['550']['std'] for v in fcdep_ts]).filled(float(-1.0e38)),
        #             missing_value=float(-1.0e38))

        # #Make some gifs at the end
        # for plttype in ['monarch-%s-*-fc.png' % (mvar),
        #                 'VIIRS_550_AODMap_*_y.png','VIIRS_550_AODMap_*_fcdep.png',
        #                 'VIIRS_550_AODMap_*_hofx.png',
        #                 'VIIRS_AODHisto_*_fcdep.png']:
        #     subprocess.run(['convert','-delay','80','-loop','0',os.path.join(framedir,plttype),
        #                     os.path.join(gifdir,plttype.replace('*','anim').replace('.png','.gif'))])
                
        # Here we can plot obs for all day if needed
        # yall.PlotObsValues(ObsOnMapOpt={'range':maprange[mvar],'markersize':markersize,
        #                             # 'colormap':'viridis'},
        #                             'colormap':'YlOrBr'},
        #                            PlotRegion=pltbounds,MapOpt={'AddGrid':False},
        #                            outNameSuffix='yall',outdir=outdir)
        # yall.PlotMaskValues(MaskOnMapOpt={'range':maprange[mvar],'markersize':markersize},
        #                            PlotRegion=pltbounds,MapOpt={'AddGrid':False},
        #                            outNameSuffix='assim_mask',outdir=outdir)

if __name__ == "__main__":
    
    parser = ArgumentParser(
      formatter_class=RawDescriptionHelpFormatter,
      description=('''\n\
      Plots results from operational DA experiments.
      E.g. ./postprocfct.py -n 1 -d "2024-07-15"
      '''))
    parser.add_argument(
      "-n", "--ndays", dest="ndays",
      help="Number of days to process")
    parser.add_argument(
      "-d", "--date", dest="date",
      help="First date to process")
    
    parser.set_defaults(ndays=2,date='2021-03-18')
    opt = parser.parse_args()
    
    loopovertime(opt)


